(* Delphi-Unit (Unicode)
   Calculate MD5 signature from Stream (160bit hash size)

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   Acknowledgements:
     based on the routines by
     Dave Barton (davebarton@bigfoot.com)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jul. 2012
   Vers. 2 - Jun. 2012 : speeded up by optimizing the code
   *)

unit SHA1;

interface

uses System.Sysutils, Winapi.Windows, System.Classes, CBFunctions;

function SHA1Stream (fs : TStream; CallBack : TProgressEvent = nil) : AnsiString;
procedure SHA1Break;

function SHA1AnsiString(const s : AnsiString) : AnsiString;
function SHA1String(const s : String) : AnsiString;

//******************************************************************************
implementation

const
  SHABufSize = 1024*1024*32;
  CtxBufSize = 64;

var
  UserBreak : boolean;

type
  TSHA1Digest= array[0..19] of byte;

  TInt64 = record
    case integer of
    0: (AsInt64 : int64);
    1: (Lo, Hi  : Cardinal);
    end;

  TSHA1Context= record
    Hash: array[0..4] of DWord;
    Length : TInt64;
//    Hi, Lo: integer;
    Buffer: array[0..CtxBufSize-1] of byte;
    Index: integer;
  end;

function LRot16(X: word; c: integer): word; assembler;
asm
  mov ecx,&c
  mov ax,&X
  rol ax,cl
  mov &Result,ax
end;

function RRot16(X: word; c: integer): word; assembler;
asm
  mov ecx,&c
  mov ax,&X
  ror ax,cl
  mov &Result,ax
end;

function LRot32(X: dword; c: integer): dword; register; assembler;
asm
  mov ecx, edx
  rol eax, cl
end;

function RRot32(X: dword; c: integer): dword; register; assembler;
asm
  mov ecx, edx
  ror eax, cl
end;

procedure XorBlock(I1, I2, O1: PByteArray; Len: integer);
var
  i: integer;
begin
  for i:= 0 to Len-1 do
    O1[i]:= I1[i] xor I2[i];
end;

procedure IncBlock(P: PByteArray; Len: integer);
begin
  Inc(P[Len-1]);
  if (P[Len-1]= 0) and (Len> 1) then
    IncBlock(P,Len-1);
end;

//******************************************************************************
function F1(x, y, z: DWord): DWord;
begin
  Result:= z xor (x and (y xor z));
end;
function F2(x, y, z: DWord): DWord;
begin
  Result:= x xor y xor z;
end;
function F3(x, y, z: DWord): DWord;
begin
  Result:= (x and y) or (z and (x or y));
end;

//******************************************************************************
function RB(A: DWord): DWord;
begin
  Result:= (A shr 24) or ((A shr 8) and $FF00) or ((A shl 8) and $FF0000) or (A shl 24);
end;

procedure SHA1Compress(var Data: TSHA1Context);
var
  A, B, C, D, E, T: DWord;
  W: array[0..79] of DWord;
  i: integer;
begin
  Move(Data.Buffer,W,Sizeof(Data.Buffer));
  for i:= 0 to 15 do
    W[i]:= RB(W[i]);
  for i:= 16 to 79 do
    W[i]:= LRot32(W[i-3] xor W[i-8] xor W[i-14] xor W[i-16],1);
  A:= Data.Hash[0]; B:= Data.Hash[1]; C:= Data.Hash[2]; D:= Data.Hash[3]; E:= Data.Hash[4];
  for i:= 0 to 19 do
  begin
    T:= LRot32(A,5) + F1(B,C,D) + E + W[i] + $5A827999;
    E:= D; D:= C; C:= LRot32(B,30); B:= A; A:= T;
  end;
  for i:= 20 to 39 do
  begin
    T:= LRot32(A,5) + F2(B,C,D) + E + W[i] + $6ED9EBA1;
    E:= D; D:= C; C:= LRot32(B,30); B:= A; A:= T;
  end;
  for i:= 40 to 59 do
  begin
    T:= LRot32(A,5) + F3(B,C,D) + E + W[i] + $8F1BBCDC;
    E:= D; D:= C; C:= LRot32(B,30); B:= A; A:= T;
  end;
  for i:= 60 to 79 do
  begin
    T:= LRot32(A,5) + F2(B,C,D) + E + W[i] + $CA62C1D6;
    E:= D; D:= C; C:= LRot32(B,30); B:= A; A:= T;
  end;
  Data.Hash[0]:= Data.Hash[0] + A;
  Data.Hash[1]:= Data.Hash[1] + B;
  Data.Hash[2]:= Data.Hash[2] + C;
  Data.Hash[3]:= Data.Hash[3] + D;
  Data.Hash[4]:= Data.Hash[4] + E;
  FillChar(W,Sizeof(W),0);
  FillChar(Data.Buffer,Sizeof(Data.Buffer),0);
end;

//******************************************************************************
procedure SHA1Init(var Context: TSHA1Context);
begin
  Context.Length.AsInt64:=0; //Hi:= 0; Context.Lo:= 0;
  Context.Index:= 0;
  FillChar(Context.Buffer,Sizeof(Context.Buffer),0);
  Context.Hash[0]:= $67452301;
  Context.Hash[1]:= $EFCDAB89;
  Context.Hash[2]:= $98BADCFE;
  Context.Hash[3]:= $10325476;
  Context.Hash[4]:= $C3D2E1F0;
end;

//******************************************************************************
procedure SHA1Update(var Context: TSHA1Context; Buffer: PByte; Len: integer);
var
  n : integer;
begin
  inc(Context.Length.AsInt64,8*Len);;
  repeat
    if Len<CtxBufSize then n:=Len else n:=CtxBufSize;
    Move(Buffer^,Context.Buffer[0],n);
    if n=CtxBufSize then SHA1Compress(Context);
    inc(Buffer,n); dec(Len,n);
    until Len<=0;
  with Context do if n=CtxBufSize then Index:=0 else Index:=n;
  end;

//******************************************************************************
procedure SHA1Final(var Context: TSHA1Context; var Digest: TSHA1Digest);
var
  n : dword;
begin
  with Context do begin
    Buffer[Context.Index]:= $80;
    if Index>= 56 then SHA1Compress(Context);
    n:=RB(Length.Hi); Move(n,Buffer[56],4);
    n:=RB(Length.Lo); Move(n,Buffer[60],4);
    end;
  SHA1Compress(Context);
  Context.Hash[0]:= RB(Context.Hash[0]);
  Context.Hash[1]:= RB(Context.Hash[1]);
  Context.Hash[2]:= RB(Context.Hash[2]);
  Context.Hash[3]:= RB(Context.Hash[3]);
  Context.Hash[4]:= RB(Context.Hash[4]);
  Move(Context.Hash,Digest,Sizeof(Digest));
  FillChar(Context,Sizeof(Context),0);
end;

//------------------------------------------------------------------------------
function SHA1SelfTest: boolean;
const
  s: string= 'abc';
  OutDigest: TSHA1Digest=
    ($a9,$99,$3e,$36,$47,$06,$81,$6a,$ba,$3e,$25,$71,$78,$50,$c2,$6c,$9c,$d0,$d8,$9d);
var
  Context: TSHA1Context;
  Digest: TSHA1Digest;
begin
  SHA1Init(Context);
  SHA1Update(Context,@s[1],length(s));
  SHA1Final(Context,Digest);
  if CompareMem(@Digest,@OutDigest,Sizeof(Digest)) then
    Result:= true
  else
    Result:= false;
end;

// other check:
// Text = 'Frank jagt im komplett verwahrlosten Taxi quer durch Bayern'
// SHA1 = d8e8ece39c437e515aa8997c1a1e94f1ed2a0e62

//------------------------------------------------------------------------------
function SHA1Stream (fs : TStream; CallBack : TProgressEvent = nil) : AnsiString;
var
  Context : TSHA1Context;
  n       : integer;
  Buffer  : array of byte;
  Digest  : TSHA1Digest;
begin
  SetLength(Buffer,SHABufSize);
  SHA1Init(Context);
  UserBreak:=false;
  with fs do begin
    if assigned(CallBack) then CallBack(ptStart,Size);
    while (Position<Size) and not UserBreak do begin
      n:=Read(Buffer[0],SHABufSize);
      if assigned(CallBack) then begin
        CallBack(ptInc,n); CallBack(ptPos,Position);
        end;
      SHA1Update(Context,@Buffer[0],n);
      end;
    if assigned(CallBack) then CallBack(ptEnd,Size);
    end;
  SHA1Final(Context,Digest);
  Result := '';
  Buffer:=nil;
  if not UserBreak then for n:=0 to length(Digest)-1 do Result:=Result+AnsiLowerCase(IntToHex(Digest[n],2));
  end;

procedure SHA1Break;
begin
  UserBreak:=true;
  end;

function SHA1AnsiString(const s : AnsiString) : AnsiString;
var
  ms : TMemoryStream;
begin
  ms:=TMemoryStream.Create;
  with ms do begin
    Write(s[1],length(s));
    Position:=0;
    Result:=SHA1Stream(ms);
    Free;
    end;
  end;

function SHA1String(const s : String) : AnsiString;
var
  ms : TMemoryStream;
begin
  ms:=TMemoryStream.Create;
  with ms do begin
    Write(s[1],2*length(s));
    Position:=0;
    Result:=SHA1Stream(ms);
    Free;
    end;
  end;

end.
